
$.fullCalendar.locale("fr", {
    buttonText: {
        year: "Année",
        month: "Mois",
        week: "Semaine",
        day: "Jour",
        list: "Mon planning",
        today: "Aujourd'hui "
    },
    allDayHtml: "Toute la<br/>journée",
    eventLimitText: "en plus",
    noEventsMessage: "Aucun événement à afficher"
});